@@ -1,5 +1,7 @@
+#include <base-logging/Logging.hpp>
 #include <iodrivers_base/Driver.hpp>
 #include <iodrivers_base/Timeout.hpp>
+#include <iodrivers_base/URI.hpp>
 
 #include <errno.h>
 #include <sys/types.h>
@@ -15,6 +17,7 @@
 #include <cstring>
 #include <sstream>
 #include <iostream>
+#include <stdexcept>
 
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -22,6 +25,7 @@
 #include <netinet/udp.h>
 #include <netdb.h>
 
+#include <boost/regex.hpp>
 #include <boost/lexical_cast.hpp>
 #include <iodrivers_base/IOStream.hpp>
 #include <iodrivers_base/IOListener.hpp>
@@ -45,6 +49,7 @@
 using namespace std;
 using base::Time;
 using namespace iodrivers_base;
+using boost::lexical_cast;
 
 string Driver::printable_com(std::string const& str)
 { return printable_com(str.c_str(), str.size()); }
@@ -154,91 +159,152 @@ int Driver::getFileDescriptor() const
 }
 bool Driver::isValid() const { return m_stream; }
 
-void Driver::openURI(std::string const& uri)
-{
-    // Modes:
-    //   0 for serial
-    //   1 for TCP
-    //   2 for UDP
-    //   3 for UDP server
-    //   4 for file (either Unix sockets or named FIFOs)
-    int mode_idx = -1;
-    char const* modes[6] = { "serial://", "tcp://", "udp://", "udpserver://", "file://", "test://" };
-    for (int i = 0; i < 6; ++i)
-    {
-        if (uri.compare(0, strlen(modes[i]), modes[i]) == 0)
-        {
-            mode_idx = i;
-            break;
+static void validateURIScheme(std::string const& scheme) {
+    char const* knownSchemes[6] = { "serial", "tcp", "udp", "udpserver", "file", "test" };
+    for (int i = 0; i < 6; ++i) {
+        if (scheme == knownSchemes[i]) {
+            return;
         }
     }
 
-    if (mode_idx == -1)
-        throw std::runtime_error("unknown URI " + uri);
+    throw std::runtime_error("unknown scheme " + scheme);
+}
 
-    string device = uri.substr(strlen(modes[mode_idx]));
-
-    // Find a :[additional_info] marker
-    string::size_type marker = device.find_last_of(":");
-    int additional_info = 0;
-    if (marker != string::npos)
-    {
-        additional_info = boost::lexical_cast<int>(device.substr(marker + 1));
-        device = device.substr(0, marker);
+/** Backward-compatibility code to handle the old syntax udp://host:remote_port:local_port
+ *
+ * It transforms it into the new udp://host:remote_port?local_port=PORT URI
+ */
+static URI backwardParseBidirectionalUDP(string const& uri_string) {
+    size_t new_style = uri_string.find_first_of("?&=");
+    if (new_style != string::npos) {
+        return URI::parse(uri_string);
     }
 
-    if (mode_idx == 0)
-    { // serial://DEVICE:baudrate
-        if (marker == string::npos)
-            throw std::runtime_error("missing baudrate specification in serial:// URI");
-        openSerial(device, additional_info);
-        return;
-    }
-    else if (mode_idx == 1)
-    { // TCP tcp://hostname:port
-        if (marker == string::npos)
-            throw std::runtime_error("missing port specification in tcp:// URI");
-        return openTCP(device, additional_info);
-    }
-    else if (mode_idx == 2)
-    { // UDP udp://hostname:remoteport
-        if (marker == string::npos)
-            throw std::runtime_error("missing port specification in udp:// URI");
+    size_t first_colon = uri_string.find_first_of(":", 6);
+    size_t last_colon = uri_string.find_last_of(":");
 
-        string::size_type remote_port_marker = device.find_last_of(":");
-        if (remote_port_marker != string::npos)
-        {
-            int remote_port = boost::lexical_cast<int>(device.substr(remote_port_marker + 1));
-            device = device.substr(0, remote_port_marker);
+    if (first_colon == last_colon) {
+        return URI::parse(uri_string);
+    }
 
-            return openUDPBidirectional(device, remote_port, additional_info);
+    string local_port = uri_string.substr(last_colon + 1, string::npos);
+    string new_uri = uri_string.substr(0, last_colon) + "?local_port=" + local_port;
+    return URI::parse(new_uri);
+}
+
+void Driver::openURI(std::string const& uri_string) {
+    URI uri;
+    if (uri_string.substr(0, 6) == "udp://") {
+        uri = backwardParseBidirectionalUDP(uri_string);
+    }
+    else {
+        uri = URI::parse(uri_string);
+    }
+    string scheme = uri.getScheme();
+    validateURIScheme(scheme);
+
+    if (scheme == "serial") { // serial://DEVICE:baudrate
+        if (uri.getPort() == 0) {
+            throw std::invalid_argument("missing baud rate specification in serial URI");
         }
-
-        return openUDP(device, additional_info);
+        openSerial(uri.getHost(), uri.getPort(), SerialConfiguration::fromURI(uri));
     }
-    else if (mode_idx == 3)
-    { // UDP udpserver://localport
-        return openUDP("", boost::lexical_cast<int>(device));
+    else if (scheme == "tcp") { // TCP tcp://hostname:port
+        if (uri.getPort() == 0) {
+            throw std::invalid_argument("missing port specification in tcp URI");
+        }
+        openTCP(uri.getHost(), uri.getPort());
     }
-    else if (mode_idx == 4)
-    { // file file://path
-        return openFile(device);
+    else if (scheme == "udp") { // UDP udp://hostname:remoteport
+        openURI_UDP(uri);
     }
-    else if (mode_idx == 5)
-    { // test://
+    else if (scheme == "udpserver") { // UDP udpserver://localport
+        openUDPServer(stoi(uri.getHost()));
+    }
+    else if (scheme == "file") { // file file://path
+        return openFile(uri.getHost());
+    }
+    else if (scheme == "test") { // test://
         if (!dynamic_cast<TestStream*>(getMainStream()))
             openTestMode();
     }
 }
 
+void Driver::openURI_UDP(URI const& uri) {
+    if (uri.getPort() == 0) {
+        throw std::invalid_argument("missing port specification in udp URI");
+    }
+
+    string local_port = uri.getOption("local_port");
+    string ignore_connrefused = uri.getOption("ignore_connrefused");
+    string ignore_hostunreach = uri.getOption("ignore_hostunreach");
+    string ignore_netunreach = uri.getOption("ignore_netunreach");
+    string connected = uri.getOption("connected");
+
+    if (local_port.empty() && ignore_connrefused.empty()) {
+        LOG_WARN_S << "udp://host:port streams historically would report connection "
+                        "refused errors. This default behavior will change in the "
+                        "future." << endl;
+        LOG_WARN_S << "Set the ignore_connrefused option to 1 to update to the new "
+                        "behavior and remove this warning, or set it to 0 to ensure "
+                        "that the behavior will be retained when the default changes"
+                    << endl;
+    }
+    if (!local_port.empty() && connected.empty()) {
+        LOG_WARN_S << "udp://host:remote_port?local_port=PORT historically was not "
+                        "connecting the socket, which means that any remote host could "
+                        "send messages to the local socket." << endl;
+        LOG_WARN_S << "This default behavior will change in the future. Set the "
+                        "connected option to 1 to update to the new behavior, that is "
+                        "allowing only the specified remote host to send packets."
+                    << endl;
+        LOG_WARN_S << "Set to 0 to keep the current behavior even after the "
+                        "default is changed " << endl;
+    }
+
+    if (connected.empty()) {
+        connected = local_port.empty() ? "1" : "0";
+    }
+
+    bool is_connected = (connected == "1");
+
+    if (ignore_connrefused.empty()) {
+        ignore_connrefused = is_connected ? (local_port.empty() ? "0" : "1") : "1";
+    }
+    if (ignore_hostunreach.empty()) {
+        ignore_hostunreach = "0";
+    }
+    if (ignore_netunreach.empty()) {
+        ignore_netunreach = "0";
+    }
+    if (local_port.empty()) {
+        local_port = "0";
+    }
+    if (ignore_connrefused == "0" && !is_connected) {
+        throw std::invalid_argument(
+            "cannot set ignore_connrefused=0 on an unconnected UDP stream"
+        );
+    }
+
+    openUDPBidirectional(uri.getHost(), uri.getPort(), stoi(local_port),
+                         ignore_connrefused == "1",
+                         connected == "1",
+                         ignore_hostunreach == "1",
+                         ignore_netunreach == "1");
+}
+
 void Driver::openTestMode()
 {
     setMainStream(new TestStream);
 }
 
-bool Driver::openSerial(std::string const& port, int baud_rate)
-{
-    setFileDescriptor(Driver::openSerialIO(port, baud_rate), true, false);
+
+
+bool Driver::openSerial(std::string const& device, int baudrate,
+                        SerialConfiguration const& configuration) {
+
+    setFileDescriptor(Driver::openSerialIO(device, baudrate), true, false);
+    setSerialConfiguration(configuration);
     return true;
 }
 
@@ -248,73 +314,102 @@ bool Driver::openInet(const char *hostname, int port)
     return true;
 }
 
-static int createIPServerSocket(int port, addrinfo const& hints)
-{
-    struct addrinfo *result;
-    string port_as_string = boost::lexical_cast<string>(port);
-    int ret = getaddrinfo(NULL, port_as_string.c_str(), &hints, &result);
-    if (ret != 0)
-        throw UnixError("cannot resolve server port " + port_as_string);
+struct AddrinfoGuard {
+    addrinfo* ptr;
+    AddrinfoGuard(addrinfo* ptr)
+        : ptr(ptr) {}
+    ~AddrinfoGuard() {
+        freeaddrinfo(ptr);
+    }
+};
 
-    int sfd = -1;
-    struct addrinfo *rp;
-    for (rp = result; rp != NULL; rp = rp->ai_next) {
-        sfd = socket(rp->ai_family, rp->ai_socktype,
-                rp->ai_protocol);
+static int createIPServerSocket(const char* port, addrinfo const& hints)
+{
+    addrinfo *candidates;
+    int ret = getaddrinfo(NULL, port, &hints, &candidates);
+    if (ret != 0)
+        throw UnixError("cannot resolve server port " + string(port));
+
+    AddrinfoGuard guard(candidates);
+
+    for (addrinfo* rp = candidates; rp != NULL; rp = rp->ai_next) {
+        int sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
         if (sfd == -1)
             continue;
 
-        if (::bind(sfd, rp->ai_addr, rp->ai_addrlen) == 0)
-            break;                  /* Success */
+        if (::bind(sfd, rp->ai_addr, rp->ai_addrlen) == 0) {
+            return sfd;
+        }
 
         ::close(sfd);
     }
-    freeaddrinfo(result);
 
-    if (rp == NULL)
-        throw UnixError("cannot open server socket on port " + port_as_string);
-
-    return sfd;
+    throw UnixError("cannot open server socket on port " + string(port));
 }
 
-static int createIPClientSocket(const char *hostname, const char *port, addrinfo const& hints, struct sockaddr *addr, size_t *addr_len)
-{
-    struct addrinfo *result;
-    int ret = getaddrinfo(hostname, port, &hints, &result);
-    if (ret != 0)
+static int createIPClientSocket(
+    const char *hostname, const char *port, addrinfo const& hints,
+    sockaddr_storage* sockaddr, size_t* sockaddr_len
+) {
+    addrinfo *candidates;
+    int ret = getaddrinfo(hostname, port, &hints, &candidates);
+    if (ret != 0) {
         throw UnixError("cannot resolve client port " + string(port));
+    }
 
-    int sfd = -1;
-    struct addrinfo *rp;
-    for (rp = result; rp != NULL; rp = rp->ai_next) {
-        sfd = socket(rp->ai_family, rp->ai_socktype,
-                rp->ai_protocol);
+    AddrinfoGuard guard(candidates);
+
+    for (addrinfo* rp = candidates; rp != NULL; rp = rp->ai_next) {
+        int sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
         if (sfd == -1)
             continue;
 
-        if (connect(sfd, rp->ai_addr, rp->ai_addrlen) == 0)
-            break;                  /* Success */
+        if (connect(sfd, rp->ai_addr, rp->ai_addrlen) != 0) {
+            ::close(sfd);
+            continue;
+        }
 
-        ::close(sfd);
+        if (sockaddr) {
+            memcpy(sockaddr, rp->ai_addr, rp->ai_addrlen);
+            *sockaddr_len = rp->ai_addrlen;
+        }
+        return sfd;
     }
 
-    if (rp == NULL)
-    {
-        freeaddrinfo(result);
-        throw UnixError("cannot open client socket on port " + string(port));
+    throw UnixError("cannot open client socket on port " + string(port));
+}
+
+static pair<sockaddr_storage, size_t> connectIPSocket(
+    int fd, const char* hostname, const char* port, addrinfo const& hints
+) {
+    addrinfo *candidates;
+    int ret = getaddrinfo(hostname, port, &hints, &candidates);
+    if (ret != 0) {
+        throw UnixError("cannot resolve client port " + string(port));
     }
 
-    if (addr != NULL) *addr = *(rp->ai_addr);
-    if (addr_len != NULL) *addr_len = rp->ai_addrlen;
+    AddrinfoGuard guard(candidates);
 
-    freeaddrinfo(result);
+    for (addrinfo* rp = candidates; rp != NULL; rp = rp->ai_next) {
+        if (connect(fd, rp->ai_addr, rp->ai_addrlen) != 0) {
+            continue;
+        }
 
-    return sfd;
+        pair<sockaddr_storage, size_t> result;
+        memcpy(&result.first, rp->ai_addr, rp->ai_addrlen);
+        result.second = rp->ai_addrlen;
+        return result;
+    }
+
+    throw UnixError("cannot connect client socket on port " + string(port));
 }
 
 void Driver::openIPClient(std::string const& hostname, int port, addrinfo const& hints)
 {
-    int sfd = createIPClientSocket(hostname.c_str(), boost::lexical_cast<string>(port).c_str(), hints, NULL, NULL);
+    int sfd = createIPClientSocket(
+        hostname.c_str(), lexical_cast<string>(port).c_str(), hints,
+        nullptr, nullptr
+    );
     setFileDescriptor(sfd);
 }
 
@@ -341,45 +436,139 @@ void Driver::openUDP(std::string const& hostname, int port)
 {
     if (hostname.empty())
     {
-        struct addrinfo hints;
-        memset(&hints, 0, sizeof(struct addrinfo));
-        hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
-        hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
-        hints.ai_flags = AI_PASSIVE;    /* For wildcard IP address */
+        LOG_WARN_S << "openUDP: providing an empty hostname is "
+                      "deprecated, use openUDPServer instead" << endl;
+        return openUDPServer(port);
+    }
 
-        int sfd = createIPServerSocket(port, hints);
-        setMainStream(new UDPServerStream(sfd,true));
-    }
-    else
-    {
-        struct addrinfo hints;
-        memset(&hints, 0, sizeof(struct addrinfo));
-        hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
-        hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
-        openIPClient(hostname, port, hints);
-    }
+    struct addrinfo hints;
+    memset(&hints, 0, sizeof(struct addrinfo));
+    hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
+    hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
+    openIPClient(hostname, port, hints);
 }
 
-void Driver::openUDPBidirectional(std::string const& hostname, int out_port, int in_port)
+void Driver::openUDPServer(int port)
 {
-    struct addrinfo in_hints;
-    memset(&in_hints, 0, sizeof(struct addrinfo));
-    in_hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
-    in_hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
-    in_hints.ai_flags = AI_PASSIVE;    /* For wildcard IP address */
+    struct addrinfo hints;
+    memset(&hints, 0, sizeof(struct addrinfo));
+    hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
+    hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
+    hints.ai_flags = AI_PASSIVE;    /* For wildcard IP address */
 
-    struct addrinfo out_hints;
-    memset(&out_hints, 0, sizeof(struct addrinfo));
-    out_hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
-    out_hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
+    int sfd = createIPServerSocket(lexical_cast<string>(port).c_str(), hints);
+    setMainStream(new UDPServerStream(sfd, true));
+}
 
-    struct sockaddr peer;
-    size_t peer_len;
-    int peerfd = createIPClientSocket(hostname.c_str(), boost::lexical_cast<string>(out_port).c_str(), out_hints, &peer, &peer_len);
-    ::close(peerfd);
+void Driver::openUDPBidirectional(
+    std::string const& hostname, int remote_port, int local_port,
+    bool ignore_connrefused,
+    bool connected,
+    bool ignore_hostunreach,
+    bool ignore_netunreach
+) {
+    struct addrinfo local_hints;
+    memset(&local_hints, 0, sizeof(struct addrinfo));
+    local_hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
+    local_hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
+    local_hints.ai_flags = AI_PASSIVE;    /* For wildcard IP address */
 
-    int sfd = createIPServerSocket(in_port, in_hints);
-    setMainStream(new UDPServerStream(sfd, true, &peer, &peer_len));
+    int sfd = createIPServerSocket(lexical_cast<string>(local_port).c_str(), local_hints);
+
+    struct addrinfo remote_hints;
+    memset(&remote_hints, 0, sizeof(struct addrinfo));
+    remote_hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
+    remote_hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
+
+    pair<sockaddr_storage, size_t> peer;
+    if (connected) {
+        peer = connectIPSocket(
+            sfd, hostname.c_str(), lexical_cast<string>(remote_port).c_str(),
+            remote_hints
+        );
+    }
+    else {
+        int fd = createIPClientSocket(
+            hostname.c_str(), lexical_cast<string>(remote_port).c_str(),
+            remote_hints, &peer.first, &peer.second
+        );
+        ::close(fd);
+    }
+    auto stream = new UDPServerStream(
+        sfd, true,
+        reinterpret_cast<sockaddr*>(&peer.first), &peer.second
+    );
+    stream->setIgnoreEconnRefused(ignore_connrefused);
+    stream->setIgnoreEhostUnreach(ignore_hostunreach);
+    stream->setIgnoreEnetUnreach(ignore_netunreach);
+    setMainStream(stream);
+}
+
+SerialConfiguration Driver::parseSerialConfiguration(std::string const &description) {
+    boost::regex ex = boost::regex(
+        "^([5-8])([neo])([12])$",
+        boost::regex_constants::icase
+    );
+
+    boost::smatch sm;
+    if (!boost::regex_match (description, sm, ex)) {
+        throw invalid_argument("Invalid serial configuration");
+    }
+
+    SerialConfiguration serial_config;
+    serial_config.byte_size = static_cast<ByteSize>(atoi(sm[1].str().c_str()));
+    serial_config.parity = static_cast<ParityChecking>(toupper(sm[2].str().at(0)));
+    serial_config.stop_bits = static_cast<StopBits>(atoi(sm[3].str().c_str()));
+
+    return serial_config;
+}
+
+
+void Driver::setSerialConfiguration(SerialConfiguration const& serial_config)
+{
+    struct termios tio;
+    int fd = getFileDescriptor();
+
+    if (tcgetattr(fd, &tio)) {
+        throw UnixError("Driver::setSerialConfiguration: Failed to get terminal info\n");
+    }
+
+    if (serial_config.parity == PARITY_NONE) {
+        tio.c_cflag &= ~PARENB;
+    } else {
+        tio.c_cflag |= PARENB;
+        if (serial_config.parity == PARITY_EVEN) {
+            tio.c_cflag &= ~PARODD;
+        } else {
+            tio.c_cflag |= PARODD;
+        }
+    }
+
+    tio.c_cflag &= ~CSIZE;
+    switch (serial_config.byte_size) {
+        case BITS_5:
+            tio.c_cflag |= CS5;
+            break;
+        case BITS_6:
+            tio.c_cflag |= CS6;
+            break;
+        case BITS_7:
+            tio.c_cflag |= CS7;
+            break;
+        case BITS_8:
+            tio.c_cflag |= CS8;
+            break;
+    }
+
+    if (serial_config.stop_bits == STOP_BITS_ONE) {
+        tio.c_cflag &= ~CSTOPB;
+    } else {
+        tio.c_cflag |= CSTOPB;
+    }
+
+    if (tcsetattr(fd, TCSANOW, &tio) != 0) {
+        throw UnixError("Driver::setSerialConfiguration: Failed to set terminal info\n");
+    }
 }
 
 int Driver::openSerialIO(std::string const& port, int baud_rate)
@@ -424,15 +613,15 @@ bool Driver::setSerialBaudrate(int fd, int brate) {
     bool custom_rate = false;
 #endif
     switch(brate) {
-	case(SERIAL_1200):
-	    tc_rate = B1200;
-	    break;
-	case(SERIAL_2400):
-	    tc_rate = B2400;
-	    break;
-	case(SERIAL_4800):
-	    tc_rate = B4800;
-	    break;
+        case(SERIAL_1200):
+            tc_rate = B1200;
+            break;
+        case(SERIAL_2400):
+            tc_rate = B2400;
+            break;
+        case(SERIAL_4800):
+            tc_rate = B4800;
+            break;
         case(SERIAL_9600):
             tc_rate = B9600;
             break;
@@ -460,10 +649,13 @@ bool Driver::setSerialBaudrate(int fd, int brate) {
         case(SERIAL_921600):
             tc_rate = B921600;
             break;
+        case(SERIAL_1000000):
+            tc_rate = B1000000;
+            break;
         default:
 #ifdef __gnu_linux__
-	    tc_rate = B38400;
-	    custom_rate = true;
+            tc_rate = B38400;
+            custom_rate = true;
             std::cerr << "Using custom baud rate " << brate << std::endl;
 #else
             std::cerr << "Non-standard baud rate selected. This is only supported on linux." << std::endl;
@@ -476,20 +668,28 @@ bool Driver::setSerialBaudrate(int fd, int brate) {
     ioctl(fd, TIOCGSERIAL, &ss);
     if( custom_rate )
     {
-	ss.flags = (ss.flags & ~ASYNC_SPD_MASK) | ASYNC_SPD_CUST;
-	ss.custom_divisor = (ss.baud_base + (brate / 2)) / brate;
-	int closestSpeed = ss.baud_base / ss.custom_divisor;
+        ss.flags = (ss.flags & ~ASYNC_SPD_MASK) | ASYNC_SPD_CUST;
+        ss.custom_divisor = (ss.baud_base + (brate / 2)) / brate;
+        
+        if (ss.custom_divisor == 0) {
+            std::cerr << "Cannot set custom serial rate to " << brate
+                << " as the calculated divisor is zero for baud_base of " << ss.baud_base << "."
+                << std::endl;
+                return false;
+        }
+        
+        int closestSpeed = ss.baud_base / ss.custom_divisor;
 
-	if (closestSpeed < brate * 98 / 100 || closestSpeed > brate * 102 / 100)
-	{
-	    std::cerr << "Cannot set custom serial rate to " << brate
-		<< ". The closest possible value is " << closestSpeed << "."
-		<< std::endl;
-	}
+        if (closestSpeed < brate * 98 / 100 || closestSpeed > brate * 102 / 100)
+        {
+            std::cerr << "Cannot set custom serial rate to " << brate
+                << ". The closest possible value is " << closestSpeed << "."
+                << std::endl;
+        }
     }
     else
     {
-	ss.flags &= ~ASYNC_SPD_MASK;
+        ss.flags &= ~ASYNC_SPD_MASK;
     }
     ioctl(fd, TIOCSSERIAL, &ss);
 #endif
@@ -532,9 +732,9 @@ std::pair<uint8_t const*, int> Driver::findPacket(uint8_t const* buffer, int buf
     // the buffer
     if( extract_result > buffer_size )
         throw length_error("extractPacket() returned result size "
-                + boost::lexical_cast<string>(extract_result)
+                + lexical_cast<string>(extract_result)
                 + ", which is larger than the buffer size "
-                + boost::lexical_cast<string>(buffer_size) + ".");
+                + lexical_cast<string>(buffer_size) + ".");
 
     if (0 == extract_result)
         return make_pair(buffer, 0);
@@ -658,10 +858,7 @@ int Driver::readRaw(uint8_t* buffer, int out_buffer_size,
     while (buffer_fill < out_buffer_size && now <= global_deadline)
     {
         auto deadline = min(global_deadline, last_char + inter_byte_timeout);
-        try {
-            m_stream->waitRead(deadline - now);
-        }
-        catch(TimeoutError&) {
+        if (!m_stream->waitRead(deadline - now)) {
             break;
         }
         int c = m_stream->read(buffer + buffer_fill,
@@ -686,7 +883,7 @@ int Driver::readRaw(uint8_t* buffer, int out_buffer_size,
 pair<int, bool> Driver::readPacketInternal(uint8_t* buffer, int out_buffer_size)
 {
     if (out_buffer_size < MAX_PACKET_SIZE)
-        throw length_error("readPacket(): provided buffer too small (got " + boost::lexical_cast<string>(out_buffer_size) + ", expected at least " + boost::lexical_cast<string>(MAX_PACKET_SIZE) + ")");
+        throw length_error("readPacket(): provided buffer too small (got " + lexical_cast<string>(out_buffer_size) + ", expected at least " + lexical_cast<string>(MAX_PACKET_SIZE) + ")");
 
     // How many packet bytes are there currently in +buffer+
     int packet_size = 0;
@@ -770,8 +967,8 @@ int Driver::readPacket(uint8_t* buffer, int buffer_size,
 {
     if (buffer_size < MAX_PACKET_SIZE) {
         throw length_error("readPacket(): provided buffer too small (got "
-                + boost::lexical_cast<string>(buffer_size) + ", expected at least "
-                + boost::lexical_cast<string>(MAX_PACKET_SIZE) + ")");
+                + lexical_cast<string>(buffer_size) + ", expected at least "
+                + lexical_cast<string>(MAX_PACKET_SIZE) + ")");
     }
 
     if (!isValid()) {
@@ -820,24 +1017,22 @@ int Driver::readPacket(uint8_t* buffer, int buffer_size,
             throw TimeoutError(
                 timeout_type,
                 "readPacket(): no data after waiting "
-                + boost::lexical_cast<string>((now - start_time).toMilliseconds()) + "ms");
+                + lexical_cast<string>((now - start_time).toMilliseconds()) + "ms");
         }
 
         // we still have time left to wait for arriving data. see how much
         Time remaining = deadline - now;
-        try {
-            // calls select and waits until a new read can be actually performed (in the next
-            // while-iteration)
-            m_stream->waitRead(remaining);
-        }
-        catch (TimeoutError& e)
+        
+        // calls select and waits until a new read can be actually performed (in the next
+        // while-iteration)
+        if (!m_stream->waitRead(remaining)) 
         {
             auto total_wait = Time::now() - start_time;
             throw TimeoutError(timeout_type,
                 "readPacket(): no data waiting for data. Last wait lasted "
-                + boost::lexical_cast<string>(remaining.toMilliseconds()) + "ms, "
-                + "out of a total wait of " +
-                boost::lexical_cast<string>(total_wait.toMilliseconds()) + "ms");
+                + lexical_cast<string>(remaining.toMilliseconds()) + "ms, "
+                + "out of a total wait of "
+                + lexical_cast<string>(total_wait.toMilliseconds()) + "ms");
         }
     }
 }
@@ -876,7 +1071,10 @@ bool Driver::writePacket(uint8_t const* buffer, int buffer_size, int timeout)
             throw TimeoutError(TimeoutError::PACKET, "writePacket(): timeout");
 
         int remaining_timeout = time_out.timeLeft();
-        m_stream->waitWrite(Time::fromMicroseconds(remaining_timeout * 1000));
+        
+        if (!m_stream->waitWrite(Time::fromMicroseconds(remaining_timeout * 1000))) {
+            throw TimeoutError(TimeoutError::NONE, "waitRead(): timeout");
+        };
     }
 }
 
@@ -885,4 +1083,4 @@ bool Driver::eof() const
     if (!m_stream)
         throw std::runtime_error("eof(): invalid stream");
     return m_stream->eof();
-}
\ No newline at end of file
+}
