@@ -5,11 +5,12 @@
 #include <stdint.h>
 #include <stdio.h>
 #include <unistd.h>
-#include <vector>
-#include <unistd.h>
-#include <iodrivers_base/Status.hpp>
-#include <iodrivers_base/Exceptions.hpp>
 #include <set>
+#include <vector>
+#include <iodrivers_base/Exceptions.hpp>
+#include <iodrivers_base/SerialConfiguration.hpp>
+#include <iodrivers_base/Status.hpp>
+#include <iodrivers_base/URI.hpp>
 
 struct addrinfo;
 
@@ -166,6 +167,12 @@ protected:
     */
     void pullBytesFromInternal(uint8_t* buffer, int skip, int size);
 
+    /** Helper for openURI to handle UDP streams
+     *
+     * They're rather complex to open because of backward compatibility reasons
+     */
+    void openURI_UDP(URI const& uri);
+
 public:
     /** Creates an Driver class for a packet-based protocol
      *
@@ -255,13 +262,23 @@ public:
     /**
     * Opens a UDP connection
     *
-    * If hostname and write port are given, the driver will be available to
-    * write data to a specified host. Otherwise, it is open in read-only mode.
+    * @param hostname the remote host to send datagrams to
+    * @param remote_port the port on the remote host
     *
-    * The read_port port can be 0 if the local port does not need to be fixed.
+    * A deprecated behavior is to give an empty hostname, which opens a UDP
+    * socket listening to the given port. The server will write to the host
+    * whose UDP packet was received last (or skips sending altogether until
+    * such a datagram was received). Use openUDPServer instead.
     */
     void openUDP(std::string const& hostname, int remote_port);
 
+    /**
+    * Opens a listening UDP socket
+    *
+    * @param port the port to bind to, set to zero to use a random port
+    */
+    void openUDPServer(int port);
+
     /**
     * Opens a UDP connection
     *
@@ -269,8 +286,13 @@ public:
     * write data to a specified host and output port. Data will be read
     * from the input port.
     */
-    void openUDPBidirectional(std::string const& hostname, int out_port, int in_port);
-
+    void openUDPBidirectional(
+        std::string const& hostname, int remote_port, int local_port,
+        bool ignore_connrefused,
+        bool connected,
+        bool ignore_hostunreach = true,
+        bool ignore_netunreach = true
+    );
 
     /** Opens a serial port and sets it up to a sane configuration.  Use
      * then setSerialBaudrate() to change the actual baudrate of the
@@ -280,7 +302,14 @@ public:
      *
      * The return value is kept here for backward compatibility only.
      */
-    bool openSerial(std::string const& port, int baudrate);
+    bool openSerial(
+        std::string const& port, int baudrate,
+        SerialConfiguration const& serial_config = SerialConfiguration()
+    );
+
+    /** @overload
+     */
+    bool openSerial(URI const& uri);
 
     /** Opens a file from a path. It can be used for read-only tests of a
      * driver, or to connect to a named FIFO or an already-created Unix socket
@@ -319,6 +348,7 @@ public:
     bool isValid() const;
 
     // In most Linux systems also baudrates up to 1152000 are defined.
+    // usually in /usr/include/asm-generic/termbits.h
     enum SERIAL_RATES
     {
         SERIAL_1200 = 1200,
@@ -333,6 +363,7 @@ public:
         SERIAL_460800 = 460800,
         SERIAL_576000 = 576000,
         SERIAL_921600 = 921600,
+        SERIAL_1000000 = 1000000,
         SERIAL_1152000 = 1152000
     };
 
@@ -510,6 +541,12 @@ public:
      */
     bool eof() const;
 
+    /** Sets serial port configuration. Must be called after openURI(...)
+     */
+    void setSerialConfiguration(SerialConfiguration const& serial_config);
+
+    SerialConfiguration parseSerialConfiguration(std::string const &description);
+
     static std::string printable_com(std::string const& buffer);
     static std::string printable_com(uint8_t const* buffer, size_t buffer_size);
     static std::string printable_com(char const* buffer, size_t buffer_size);
