@@ -9,32 +9,32 @@
 namespace iodrivers_base {
     template<typename Driver>
     class Fixture<Driver>::BoostMockContext
-    { 
+    {
     public:
         Fixture* fixture;
         BoostMockContext(Fixture* fixture): fixture(fixture)
         {
             fixture->setMockMode(true);
         }
-        
+
         void tearDown()
         {
             try
             {
                 fixture->validateExpectationsAreEmpty();
             }
-            catch(TestEndsWithExpectationsLeftException e)
+            catch(TestEndsWithExpectationsLeftException const& e)
             {
                 BOOST_ERROR("IODRIVERS_BASE_MOCK Error: Test reached its end without satisfying all expecations.");
             }
         }
-        
+
         ~BoostMockContext()
         {
             tearDown();
             fixture->setMockMode(false);
         }
-        
+
     };
 }
 
