@@ -20,13 +20,15 @@ namespace iodrivers_base
     {
     public:
         virtual ~IOStream();
-        virtual void waitRead(base::Time const& timeout) = 0;
-        virtual void waitWrite(base::Time const& timeout) = 0;
+        virtual bool waitRead(base::Time const& timeout) = 0;
+        virtual bool waitWrite(base::Time const& timeout) = 0;
         virtual size_t read(uint8_t* buffer, size_t buffer_size) = 0;
         virtual size_t write(uint8_t const* buffer, size_t buffer_size) = 0;
         virtual void clear() = 0;
 
         virtual bool eof() const;
+        virtual bool hasIO(base::Time const& timeout);
+        virtual bool hasIO();
 
         /** If this IOStream is attached to a file descriptor, return it. Otherwise,
          * returns INVALID_FD;
@@ -51,8 +53,8 @@ namespace iodrivers_base
 
         FDStream(int fd, bool auto_close, bool has_eof = true);
         virtual ~FDStream();
-        virtual void waitRead(base::Time const& timeout);
-        virtual void waitWrite(base::Time const& timeout);
+        virtual bool waitRead(base::Time const& timeout);
+        virtual bool waitWrite(base::Time const& timeout);
         virtual size_t read(uint8_t* buffer, size_t buffer_size);
         virtual size_t write(uint8_t const* buffer, size_t buffer_size);
         virtual void clear();
@@ -64,6 +66,8 @@ namespace iodrivers_base
         bool setNonBlockingFlag(int fd);
 
         virtual int getFileDescriptor() const;
+
+        void setAutoClose(bool flag);
     };
 
     class UDPServerStream : public FDStream
@@ -73,11 +77,32 @@ namespace iodrivers_base
         UDPServerStream(int fd, bool auto_close, struct sockaddr *si_other, size_t *s_len);
         virtual size_t read(uint8_t* buffer, size_t buffer_size);
         virtual size_t write(uint8_t const* buffer, size_t buffer_size);
+        void setIgnoreEconnRefused(bool enable);
+        void setIgnoreEhostUnreach(bool enable);
+        void setIgnoreEnetUnreach(bool enable);
+
+        bool waitRead(base::Time const& timeout);
+
     protected:
+        /** Internal implementation of recvfrom to allow for mocking */
+        virtual std::pair<ssize_t, int> recvfrom(
+            uint8_t* buffer, size_t buffer_size, int flags,
+            sockaddr* s_other, socklen_t* s_len
+        );
+        /** Internal implementation of recvfrom to allow for mocking */
+        virtual std::pair<ssize_t, int> sendto(
+            uint8_t const* buffer, size_t buffer_size
+        );
+
         sockaddr m_si_other;
         unsigned int m_s_len;
         bool m_si_other_dynamic;
         bool m_has_other;
+        bool m_ignore_econnrefused;
+        bool m_ignore_ehostunreach;
+        bool m_ignore_enetunreach;
+
+        int m_wait_read_error;
     };
 }
 
