@@ -46,8 +46,8 @@ namespace iodrivers_base
          * @param reply reply message to be received by the driver
          */
         void EXPECT_REPLY(std::vector<uint8_t> const& expectation, std::vector<uint8_t> const& reply);
-        void waitRead(base::Time const& timeout);
-        void waitWrite(base::Time const& timeout);
+        bool waitRead(base::Time const& timeout);
+        bool waitWrite(base::Time const& timeout);
         size_t read(uint8_t* buffer, size_t buffer_size);
         size_t write(uint8_t const* buffer, size_t buffer_size);
         void clear();
